#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import math
import os.path
import pathlib

from typing import Tuple

import attrs

from loguru import logger
from pyspark.sql import DataFrame
from pyspark.sql.functions import lit
from pyspark.sql.types import DateType

from pramen_py.metastore.writer_base import MetastoreWriterBase
from pramen_py.models import MetastoreTable, TableFormat
from pramen_py.models.utils import get_metastore_table
from pramen_py.utils import convert_date_to_str


@attrs.define(auto_attribs=True, slots=True)
class MetastoreWriter(MetastoreWriterBase):
    """Adds writing capabilities to the Metastore.

    Makes it possible to refer tables based on it name and configurations.
    """

    def write(
        self,
        table_name: str,
        df: DataFrame,
    ) -> None:
        """Write the given table in accordance with the configuration."""

        logger.info(f"Writing table {table_name} started")

        target_table = get_metastore_table(table_name, self.tables)
        if target_table.format == TableFormat.parquet:
            save_path, count_df_items = self._write_parquet_format_table(
                df, target_table
            )
        elif target_table.format == TableFormat.delta:
            save_path, count_df_items = self._write_delta_format_table(
                df, target_table
            )
        else:
            raise NotImplementedError
        logger.info(
            f"Successfully written {count_df_items} items to {table_name} at "
            f" {save_path}"
        )

    def _write_parquet_format_table(
        self, df: DataFrame, metastore_table: MetastoreTable
    ) -> Tuple[str, int]:
        target_path = os.path.join(
            metastore_table.path,
            f"{metastore_table.info_date_settings.column}={self.info_date}",
        )
        target_path = pathlib.Path(target_path).as_posix()
        df_dropped = df.drop(metastore_table.info_date_settings.column)
        df_repartitioned, count_df_items = self._apply_repartitioning(
            df_dropped, metastore_table.records_per_partition
        )
        df_repartitioned.write.options(
            **metastore_table.writer_options
        ).format("parquet").mode("overwrite").save(target_path)
        return target_path, count_df_items

    def _write_delta_format_table(
        self, df: DataFrame, metastore_table: MetastoreTable
    ) -> Tuple[str, int]:
        info_date_str = convert_date_to_str(
            self.info_date, fmt=metastore_table.info_date_settings.format
        )

        df_with_info_date = df.withColumn(
            metastore_table.info_date_settings.column,
            lit(info_date_str).cast(DateType()),
        )
        df_repartitioned, count_df_items = self._apply_repartitioning(
            df_with_info_date, metastore_table.records_per_partition
        )
        df_writer = (
            df_repartitioned.write.format("delta")
            .mode("overwrite")
            .partitionBy(metastore_table.info_date_settings.column)
            .option("mergeSchema", "true")
            .option(
                "replaceWhere",
                f"{metastore_table.info_date_settings.column}='{info_date_str}'",
            )
            .options(**metastore_table.writer_options)
        )
        if metastore_table.path:
            df_writer.save(metastore_table.path)
            return metastore_table.path, count_df_items
        elif metastore_table.table:
            df_writer.saveAsTable(metastore_table.table)
            return metastore_table.table, count_df_items
        else:
            raise NotImplementedError

    def _apply_repartitioning(
        self, df: DataFrame, records_per_partition: int
    ) -> Tuple[DataFrame, int]:
        count_df_items = df.count()
        if records_per_partition > 0:
            num_partitions = int(
                max(1, math.ceil(count_df_items / records_per_partition))
            )
            return df.repartition(num_partitions), count_df_items
        else:
            return df, count_df_items



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pathlib

from functools import partial
from random import random
from typing import Callable, Generator, Optional, Type, TypeVar, cast

import attrs
import click

from loguru import logger
from py4j.protocol import Py4JJavaError  # type: ignore

from pramen_py import MetastoreReader, MetastoreWriter
from pramen_py.app import env
from pramen_py.models import TransformationConfig, serialization_callback
from pramen_py.runner.runner_base import (
    CLI_CALLBACK,
    Runner,
    get_current_transformer_cfg,
)
from pramen_py.transformation.transformation_base import (
    T_EXTRA_OPTIONS,
    Transformation,
)
from pramen_py.utils import (
    convert_str_to_date,
    coro,
    get_or_create_spark_session,
    load_modules,
    run_and_retry,
)


T_TRANSFORMATION = TypeVar("T_TRANSFORMATION", bound=Transformation)
DEFAULT_TRANSFORMATIONS_DIR = pathlib.Path("./transformations")
TRANSFORMATIONS_DIR_KEY_NAME = "PRAMENPY_TRANSFORMATIONS_NAMESPACE"


def overwrite_info_dates(
    config: TransformationConfig,
    info_date: str,
) -> TransformationConfig:
    """Force update the info_date in transformation configs."""
    info_date_ = convert_str_to_date(info_date)
    for transformer in config.run_transformers:
        if transformer.info_date != info_date_:
            logger.warning(
                f"info_date which is provided via cli overwrites the date "
                f"in the transformer {transformer.name}:\n"
                f"{transformer.info_date} -> {info_date_}"
            )
            object.__setattr__(transformer, "info_date", info_date_)
    return config


def get_transformations_dir() -> pathlib.Path:
    return env.path(TRANSFORMATIONS_DIR_KEY_NAME, DEFAULT_TRANSFORMATIONS_DIR)  # type: ignore


def discover_transformations() -> (
    Generator[Type[T_TRANSFORMATION], None, None]
):
    """Return generator of Transformation subclasses.

    In order to provide a set of subclasses we keep track already
    provided class names and do not allow to yield a class with the same
    name more than one time.
    """
    load_modules(get_transformations_dir())
    _transformations_names = []
    for cls in Transformation.__subclasses__():
        if cls.__name__ not in _transformations_names:
            _transformations_names.append(cls.__name__)
            yield cast(Type[T_TRANSFORMATION], cls)


@attrs.define(auto_attribs=True, slots=True)
class TransformationsRunner(Runner):
    """Impl of the Runner of the Transformation.

    This runner has the following functionalities:
    - Adds cli cmd for each transformation, with the predefined set of options
        (--config and --info-date). It is also allows to forward
        any other options which are provided via cli_options attribute of the
        Transformation base class
    - Retry running the transformer in case of a failure (see the
        pramen-py list-configuration-options)
    - Transformation command executed asynchronously with the dedicated event
        loop
    """

    def activate(self) -> None:
        self.create_cli_cmd_for_transformations()

    def create_cli_cmd_for_transformations(self) -> None:
        for T in discover_transformations():  # type: ignore
            self.cli.add_command(
                click.Command(
                    name=T.__name__,
                    callback=self.create_cli_cmd_callback(T),
                    params=[
                        click.Option(
                            ("-c", "--config"),
                            help="Transformation config file",
                            required=True,
                            callback=serialization_callback,
                            type=click.Path(exists=True),
                        ),
                        click.Option(
                            ("--info-date",),
                            help=(
                                "Information date. If specified, takes precedence "
                                "over the config values"
                            ),
                        ),
                        *T.cli_options,
                    ],
                    help=T.run.__doc__,
                )
            )

    def create_cli_cmd_callback(
        self,
        T: Type[T_TRANSFORMATION],
    ) -> Callable[CLI_CALLBACK, None]:
        async def t_run_wrapper(
            ctx: click.Context,
            config: TransformationConfig,
            info_date: Optional[str],
            **kwargs: T_EXTRA_OPTIONS,
        ) -> None:
            number_of_trials = env.int(
                "PRAMENPY_MAX_RETRIES_EXECUTE_TRANSFORMATION",
                1,
            )
            with logger.contextualize(
                transformer=f"{T.__name__}@{int(random() * 10000)}"
            ):
                await run_and_retry(
                    partial(  # type: ignore
                        self.process_transformer,
                        T=T,
                        config=config,
                        info_date=info_date,
                        kwargs=kwargs,
                    ),
                    number_of_trials=number_of_trials,
                    exception=Py4JJavaError,
                    retry_hook=self.process_transformer_retry_hook,
                )

        return cast(
            Callable[CLI_CALLBACK, None],
            click.pass_context(coro(t_run_wrapper)),
        )

    async def process_transformer(
        self,
        T: Type[T_TRANSFORMATION],
        config: TransformationConfig,
        info_date: Optional[str],
        **kwargs: T_EXTRA_OPTIONS,
    ) -> None:
        await super().process_transformer(
            T,
            config,
            info_date,
            **kwargs,
        )

        t = T(
            spark=self.spark,
            config=config,
        )

        logger.info("Executing transformer")
        t_cfg = get_current_transformer_cfg(T, config)
        if info_date:
            config = overwrite_info_dates(config, info_date)

        metastore_reader = MetastoreReader(
            spark=self.spark,
            tables=config.metastore_tables,
            info_date=t_cfg.info_date,
        )
        metastore_writer = MetastoreWriter(
            spark=self.spark,
            tables=config.metastore_tables,
            info_date=t_cfg.info_date,
        )

        transformer_result = await t.run(
            metastore_reader,
            t_cfg.info_date,
            t_cfg.options,
            **kwargs,
        )
        metastore_writer.write(
            t_cfg.output_table,
            df=transformer_result,
        )
        logger.info("Transformer executed successfully")

    async def process_transformer_retry_hook(self) -> None:
        self.spark = get_or_create_spark_session(
            env,
            force_recreate=True,
        )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime

from typing import Dict

from pyspark.sql import DataFrame
from pyspark.sql.functions import rand

from pramen_py import T_EXTRA_OPTIONS, MetastoreReader, Transformation


class IdentityTransformer(Transformation):
    async def run(
        self,
        metastore: MetastoreReader,
        info_date: datetime.date,
        options: Dict[str, str],
        **kwargs: T_EXTRA_OPTIONS,
    ) -> DataFrame:
        try:
            table_name = options["table"]
        except KeyError:
            raise KeyError(
                f"Expected 'table' key in options. Received options: {options}"
            )
        # just as example
        df = metastore.get_table(table_name)

        empty_allowed = options.get("empty.allowed", "true").lower() == "true"

        assert (
            empty_allowed or df.count() != 0
        ), f"Empty table {table_name} received"

        return df.withColumn("transform_id", rand())



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc
import datetime

from typing import ClassVar, Dict, List, TypeVar

import attrs
import click

from pyspark.sql import DataFrame, SparkSession

from pramen_py import MetastoreReader
from pramen_py.models import TransformationConfig


T_EXTRA_OPTIONS = TypeVar("T_EXTRA_OPTIONS")


@attrs.define(auto_attribs=True, slots=True)
class Transformation(metaclass=abc.ABCMeta):
    """Base transformation class.

    All Transformations should implement it and its abc members.
    """

    cli_options: ClassVar[List[click.Parameter]] = []
    spark: SparkSession = attrs.field()
    config: TransformationConfig = attrs.field()

    @abc.abstractmethod
    async def run(
        self,
        metastore: MetastoreReader,
        info_date: datetime.date,
        options: Dict[str, str],
        **kwargs: T_EXTRA_OPTIONS,
    ) -> DataFrame:
        """Abstract method for executing the transformation.

        :param metastore: instance of the MetastoreReader for convenient
            access of the tables
        :param info_date: information date for which the transformer will be
            run
        :param kwargs: if cls.cli_options contains any click.Options, then
            values from these options will be passed here
        :param options: optional settings of the transformer

        Method docstrings will be used as cli command help string.
        """
